from __future__ import annotations

import concurrent.futures
import functools
import io
import os
from collections.abc import Iterable

import npc_lims
import npc_session
import numpy as np
import numpy.typing as npt
import pandas as pd
import polars as pl

import npc_sessions.utils as utils


@functools.cache
def get_units_electrodes(
    session: str, sorting_method="codeocean_ks25", annotation_method="tissuecyte"
) -> pd.DataFrame:
    if sorting_method == "codeocean_ks25":
        units_path = npc_lims.get_units_codeoean_kilosort_path_from_s3(session)
        units = pd.read_csv(
            units_path,
            storage_options={
                "key": os.getenv("AWS_ACCESS_KEY_ID"),
                "secret": os.getenv("AWS_SECRET_ACCESS_KEY"),
            },
        )
    if annotation_method == "tissuecyte":
        units = merge_units_electrodes(session, units)

    units.drop(columns=["electrodes"], inplace=True)
    return units


def merge_units_electrodes(session: str, units: pd.DataFrame) -> pd.DataFrame:
    try:
        electrodes = utils.get_tissuecyte_electrodes_table(session)
        units = units.merge(
            electrodes,
            left_on=["group_name", "peak_channel"],
            right_on=["electrode_group_name", "channel"],
        )
        units.drop(columns=["channel"], inplace=True)
    except FileNotFoundError as e:
        print(str(e) + ". Returning units without electrodes")

    return units


def bin_spike_times(
    spike_times: npt.NDArray[np.float64], bin_time: int
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    spike_times = np.concatenate(spike_times, axis=0)  # flatten array
    return np.histogram(spike_times, bins=np.arange(0, np.max(spike_times), bin_time))


@functools.cache
def get_unit_spike_times_dict(
    session: str, unit_ids: tuple[str, ...], sorting_method="codeocean_ks25"
) -> dict[str, npt.NDArray[np.float64]]:
    # change function call depending on sorting_method
    spike_times_dict = {}
    if sorting_method == "codeocean_ks25":
        spike_times_path = npc_lims.get_spike_times_codeocean_kilosort_path_from_s3(
            session
        )
        with io.BytesIO(spike_times_path.read_bytes()) as f:
            spike_times = np.load(f, allow_pickle=True)

        for i in range(len(unit_ids)):
            spike_times_dict[unit_ids[i]] = spike_times[i]

    return spike_times_dict


def get_units_electrodes_spike_times(session: str, *args, **kwargs) -> pl.DataFrame:
    units_df = get_units_electrodes(session, *args, **kwargs)
    unit_ids = units_df["unit_name"].to_list()
    spike_times_dict = get_unit_spike_times_dict(
        session, tuple(unit_ids), *args, **kwargs
    )
    spike_times_df = pl.DataFrame(
        {
            "unit_name": tuple(spike_times_dict.keys()),
            "spike_times": tuple(t for t in spike_times_dict.values()),
        }
    )
    units = pl.DataFrame(units_df).join(spike_times_df, on="unit_name")
    return (
        units.with_columns(
            pl.col("device_name").str.replace("P", "p").alias("device_name"),
            pl.concat_str(
                pl.col("device_name").str.replace("Probe", f"{session}_"),
                pl.col("ks_unit_id").cast(pl.Int64),
                separator="_",
            ).alias("unit_id"),
        ).select(
            pl.exclude("unit_name", "id"),
        )
        # .filter(
        #    "default_qc",
        # )
    )


def get_aligned_spike_times(
    spike_times: npt.NDArray[np.floating],
    device_timing_on_sync: utils.EphysTimingInfoOnSync,
) -> npt.NDArray[np.float64]:
    return (
        spike_times / device_timing_on_sync.sampling_rate
    ) + device_timing_on_sync.start_time


def get_closest_channel(
    unit_location: npt.NDArray[np.float64], channel_positions: npt.NDArray[np.floating]
) -> int:
    distances = np.sum((channel_positions - unit_location) ** 2, axis=1)
    return int(np.argmin(distances))


def get_peak_channels(
    units_locations: npt.NDArray[np.float64],
    electrode_positions: npt.NDArray[np.floating],
) -> list[int]:
    peak_channels: list[int] = []

    for location in units_locations:
        unit_location = np.array([location[0], location[1]])
        peak_channels.append(get_closest_channel(unit_location, electrode_positions))

    return peak_channels


def get_amplitudes_mean_waveforms_ks25(
    templates: npt.NDArray[np.floating], ks_unit_ids: npt.NDArray[np.int64]
) -> tuple[list[float], list[npt.NDArray[np.floating]]]:
    unit_amplitudes: list[float] = []
    templates_mean: list[npt.NDArray[np.floating]] = []

    """
    @property
    def nbefore(self) -> int:
        nbefore = int(self._params["ms_before"] * self.sampling_frequency / 1000.0)
        return nbefore
    """
    before = 90  # from spike interface, ms_before = 3, # TODO: #38 @arjunsridhar12345 look at this further
    for index, _ks_unit_id in enumerate(ks_unit_ids):
        template = templates[index, :, :]
        values = -template[before, :]
        unit_amplitudes.append(np.max(values))
        templates_mean.append(template)

    return unit_amplitudes, templates_mean


def get_ampltiudes_std_waveforms_ks25(
    templates_std: npt.NDArray[np.floating], ks_unit_ids: list[int]
) -> list[npt.NDArray[np.floating]]:
    unit_templates_std: list[npt.NDArray[np.floating]] = []
    for index, _ks_unit_id in enumerate(ks_unit_ids):
        template = templates_std[index, :, :]
        unit_templates_std.append(template)

    return unit_templates_std


def get_units_spike_times_ks25(
    sorting_cached: dict[str, npt.NDArray],
    spike_times_aligned: npt.NDArray[np.float64],
    ks_unit_ids: npt.NDArray[np.int64],
) -> list[npt.NDArray[np.float64]]:
    units_spike_times: list[npt.NDArray[np.float64]] = []

    spike_labels = sorting_cached["spike_labels_seg0"]

    for ks_unit_id in ks_unit_ids:
        label_indices = np.argwhere(spike_labels == ks_unit_id)
        unit_spike_time = spike_times_aligned[label_indices].flatten()
        units_spike_times.append(unit_spike_time)

    return units_spike_times


def _device_helper(
    device_timing_on_sync: utils.EphysTimingInfoOnSync,
    spike_interface_data: utils.SpikeInterfaceKS25Data,
) -> pd.DataFrame:
    electrode_group_name = npc_session.ProbeRecord(device_timing_on_sync.device.name)
    electrode_positions = spike_interface_data.electrode_locations_xy(
        electrode_group_name
    )

    df_device_metrics = spike_interface_data.quality_metrics_df(
        electrode_group_name
    ).merge(
        spike_interface_data.template_metrics_df(electrode_group_name),
        left_index=True,
        right_index=True,
    )
    df_device_metrics["peak_channel"] = get_peak_channels(
        spike_interface_data.unit_locations(electrode_group_name),
        electrode_positions,
    )
    df_device_metrics["electrode_group_name"] = [str(electrode_group_name)] * len(
        df_device_metrics
    )

    amplitudes, mean_waveforms = get_amplitudes_mean_waveforms_ks25(
        spike_interface_data.templates_average(electrode_group_name),
        df_device_metrics.index.values,
    )

    spike_times_aligned = get_aligned_spike_times(
        spike_interface_data.sorting_cached(electrode_group_name)["spike_indexes_seg0"],
        device_timing_on_sync,
    )
    unit_spike_times = get_units_spike_times_ks25(
        spike_interface_data.sorting_cached(electrode_group_name),
        spike_times_aligned,
        df_device_metrics.index.values,  # TODO #37 @arjunsridhar12345 is this safe?
    )

    df_device_metrics["default_qc"] = spike_interface_data.default_qc(
        electrode_group_name
    )
    df_device_metrics["amplitude"] = amplitudes
    df_device_metrics["waveform_mean"] = mean_waveforms
    df_device_metrics["waveform_std"] = get_ampltiudes_std_waveforms_ks25(
        spike_interface_data.templates_std(electrode_group_name),
        df_device_metrics.index.to_list(),
    )
    df_device_metrics["spike_times"] = unit_spike_times
    df_device_metrics["unit_id"] = df_device_metrics.index.to_list()

    return df_device_metrics


def make_units_table_from_spike_interface_ks25(
    session_or_spikeinterface_data_or_path: str
    | npc_session.SessionRecord
    | utils.PathLike
    | utils.SpikeInterfaceKS25Data,
    devices_timing: Iterable[utils.EphysTimingInfoOnSync],
) -> pd.DataFrame:
    """
    >>> devices_timing = utils.get_ephys_timing_on_sync(npc_lims.get_h5_sync_from_s3('662892_20230821'), npc_lims.get_recording_dirs_experiment_path_from_s3('662892_20230821'))
    >>> units = make_units_table_from_spike_interface_ks25('662892_20230821', devices_timing)
    >>> len(units[units['electrode_group_name'] == 'probeA'])
    237
    """
    spike_interface_data = utils.get_spikeinterface_data(
        session_or_spikeinterface_data_or_path
    )

    devices_timing = tuple(
        timing for timing in devices_timing if timing.device.name.endswith("-AP")
    )

    device_to_future: dict[str, concurrent.futures.Future] = {}
    with concurrent.futures.ThreadPoolExecutor() as executor:
        for device_timing_on_sync in devices_timing:
            device_to_future[device_timing_on_sync.device.name] = executor.submit(
                _device_helper, device_timing_on_sync, spike_interface_data
            )

    return pd.concat(
        device_to_future[device].result() for device in sorted(device_to_future.keys())
    )


def format_unit_ids(
    units: pd.DataFrame, session: str | npc_session.SessionRecord
) -> pd.DataFrame:
    """Add session and probe letter"""
    units["unit_id"] = [
        f"{session}_{row.electrode_group_name.replace('probe', '')}-{row.unit_id}"
        if session not in str(row.unit_id)  # in case we aready ran this fn
        else row.unit_id
        for _, row in units.iterrows()
    ]
    return units


if __name__ == "__main__":
    import doctest

    import dotenv

    dotenv.load_dotenv(dotenv.find_dotenv(usecwd=True))
    doctest.testmod(
        optionflags=(doctest.IGNORE_EXCEPTION_DETAIL | doctest.NORMALIZE_WHITESPACE)
    )



from __future__ import annotations

import collections.abc
import contextlib
import pathlib
from collections.abc import Iterable, Iterator
from typing import Literal, SupportsFloat, TypeVar

import npc_session
import numpy as np
import numpy.typing as npt
import upath


def is_stim_file(
    path: str | upath.UPath | pathlib.Path,
    subject_spec: str | int | npc_session.SubjectRecord | None = None,
    date_spec: str | npc_session.DateRecord | None = None,
    time_spec: str | npc_session.TimeRecord | None = None,
) -> bool:
    """Does the string or path match a known stimulus file pattern.

    Optional arguments can be used to check the subject, date, and time.
    >>> is_stim_file("366122_20230101_120000.hdf5")
    True
    >>> is_stim_file("366122_20230101_120000.hdf5", date_spec="20230101")
    True
    >>> is_stim_file("366122.stim.pkl")
    True
    >>> is_stim_file("366123.stim.pkl", subject_spec="366122")
    False
    """
    path = pathlib.Path(path)
    subject_from_path = npc_session.extract_subject(path.stem)
    date_from_path = npc_session.extract_isoformat_date(path.stem)
    time_from_path = npc_session.extract_isoformat_time(path.stem)

    is_stim = False
    is_stim |= path.suffix in (".pkl",) and any(
        label in path.stem for label in ("stim", "mapping", "opto", "behavior")
    )
    is_stim |= (
        path.suffix in (".hdf5",) and bool(subject_from_path) and bool(date_from_path)
    )
    is_correct = True
    if subject_spec:
        is_correct &= subject_from_path == npc_session.SubjectRecord(str(subject_spec))
    if date_spec:
        is_correct &= date_from_path == npc_session.DateRecord(date_spec)
    if time_spec:
        is_correct &= time_from_path == npc_session.TimeRecord(time_spec)
    return is_stim and is_correct


def extract_camera_name(path: str) -> Literal["eye", "face", "behavior"]:
    names: dict[str, Literal["eye", "face", "behavior"]] = {
        "eye": "eye",
        "face": "face",
        "beh": "behavior",
    }
    try:
        return names[next(n for n in names if n in str(path).lower())]
    except StopIteration as exc:
        raise ValueError(f"Could not extract camera name from {path}") from exc


def safe_index(
    array: npt.ArrayLike, indices: SupportsFloat | Iterable[SupportsFloat]
) -> npt.NDArray:
    """Checks `indices` can be safely used as array indices (i.e. all
    numerical float values are integers), then indexes into `array` using `np.where`.

    - returns nans where `indices` is nan
    - returns a scalar if `indices` is a scalar #TODO current type annotation is insufficient

    >>> safe_index([1, 2], 0)
    1
    >>> safe_index([1., 2.], 0)
    1.0
    >>> safe_index([1., 2.], np.nan)
    nan
    >>> safe_index([1., 2., 3.1], [0, np.nan, 2.0])
    array([1. , nan, 3.1])

    Type of array is preserved, if possible:
    >>> safe_index([1, 2, 3], [0., 1., 2.])
    array([1, 2, 3])

    Type of array can't be preserved if any indices are nan:
    >>> safe_index([1, 2, 3], [0, np.nan, 2.0])
    array([ 1., nan,  3.])
    """
    idx: npt.NDArray = np.array(indices)  # copy
    if not all(idx[~np.isnan(idx)] == idx[~np.isnan(idx)].astype(np.int32)):
        raise TypeError(
            f"Non-integer numerical values cannot be used as indices: {idx[np.isnan(idx)][0]}"
        )
    array = np.array(array)  # copy/make sure array can be fancy-indexed
    int_idx = np.where(np.isnan(idx), -1, idx)
    result = np.where(np.isnan(idx), np.nan, array[int_idx.astype(np.int32)])
    # np.where casts indexed array to floats just because of the
    # possibility of nans being in result, even if they aren't:
    # cast back if appropriate
    if not np.isnan(result).any():
        result = result.astype(array.dtype)
    # if indices was a scalar, return a scalar instead of a 0d array
    if not isinstance(indices, Iterable):
        assert result.size == 1
        return result.item()
    return result


K = TypeVar("K")
V = TypeVar("V")


class LazyDict(collections.abc.Mapping[K, V]):
    """Dict for postponed evaluation of functions and caching of results.

    Assign values as a tuple of (callable, args, kwargs). The callable will be
    evaluated when the key is first accessed. The result will be cached and
    returned directly on subsequent access.

    Effectively immutable after initialization.

    Initialize with a dict:
    >>> d = LazyDict({'a': (lambda x: x + 1, (1,), {})})
    >>> d['a']
    2

    or with keyword arguments:
    >>> d = LazyDict(b=(min, (1, 2), {}))
    >>> d['b']
    1
    """

    def __init__(self, *args, **kwargs) -> None:
        self._raw_dict = dict(*args, **kwargs)

    def __getitem__(self, key) -> V:
        with contextlib.suppress(TypeError):
            func, args, *kwargs = self._raw_dict.__getitem__(key)
            self._raw_dict.__setitem__(key, func(*args, **kwargs[0]))
        return self._raw_dict.__getitem__(key)

    def __iter__(self) -> Iterator[K]:
        return iter(self._raw_dict)

    def __len__(self) -> int:
        return len(self._raw_dict)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(keys={list(self._raw_dict.keys())})"


if __name__ == "__main__":
    import doctest

    import dotenv

    dotenv.load_dotenv(dotenv.find_dotenv(usecwd=True))
    doctest.testmod(
        optionflags=(doctest.IGNORE_EXCEPTION_DETAIL | doctest.NORMALIZE_WHITESPACE)
    )



from __future__ import annotations

import functools

import npc_lims
import npc_session
import numpy as np
import pandas as pd
import upath

S3_ELECTRODE_PATH = upath.UPath(
    "s3://aind-scratch-data/arjun.sridhar/tissuecyte_cloud_processed"
)

TISSUECYTE_MICRONS_PER_PIXEL = 25


@functools.cache
def get_tissuecyte_annotation_files_from_s3(
    session: str | npc_session.SessionRecord,
) -> tuple[upath.UPath, ...]:
    """For each probe inserted, get a csv file containing CCF coordinates for each
    electrode (channel) on the probe.

    >>> electrode_files = get_tissuecyte_annotation_files_from_s3('626791_2022-08-16')
    >>> assert len(electrode_files) > 0
    >>> electrode_files[0].name
    'Probe_A2_channels_626791_warped_processed.csv'
    """
    session = npc_session.SessionRecord(session)
    day = npc_lims.get_session_info(session).day
    subject_electrode_network_path = S3_ELECTRODE_PATH / str(session.subject.id)

    if not subject_electrode_network_path.exists():
        raise FileNotFoundError(
            f"CCF annotations for {session} have not been uploaded to s3"
        )

    electrode_files = tuple(
        subject_electrode_network_path.glob(
            f"Probe_*{day}_channels_{str(session.subject.id)}_warped_processed.csv"
        )
    )
    if not electrode_files:
        raise FileNotFoundError(
            f"{subject_electrode_network_path} exists, but no CCF annotation files found matching {day} and {session.subject.id} - check session day"
        )

    return electrode_files


@functools.cache
def get_tissuecyte_electrodes_table(
    session: str | npc_session.SessionRecord,
) -> pd.DataFrame:
    """Get annotation data for each electrode (channel) on each probe inserted in
    a session. Column names are ready for insertion into nwb ElectrodeTable.

    >>> df = get_tissuecyte_electrodes_table('626791_2022-08-16')
    >>> df.columns
    Index(['group_name', 'channel', 'location', 'structure', 'x', 'y', 'z'], dtype='object')
    """
    electrode_files = get_tissuecyte_annotation_files_from_s3(session)

    session_electrodes = pd.DataFrame()

    for electrode_file in electrode_files:
        probe_electrodes = pd.read_csv(electrode_file)

        probe_name = npc_session.ProbeRecord(electrode_file.stem)

        probe_electrodes["group_name"] = [str(probe_name)] * len(probe_electrodes)

        session_electrodes = pd.concat([session_electrodes, probe_electrodes])

    session_electrodes.rename(
        columns={
            "AP": "x",
            "DV": "y",
            "ML": "z",
            "region": "location",
            "region_stripped": "structure",
        },
        inplace=True,
    )
    for column in ("x", "y", "z"):
        # -1 is code for "not inserted": make this NaN
        session_electrodes[column] = session_electrodes[column].replace(-1, np.nan)
        session_electrodes[column] *= TISSUECYTE_MICRONS_PER_PIXEL
    session_electrodes = session_electrodes[
        ["group_name", "channel", "location", "structure", "x", "y", "z"]
    ]
    return session_electrodes


if __name__ == "__main__":
    import doctest

    import dotenv

    dotenv.load_dotenv(dotenv.find_dotenv(usecwd=True))
    doctest.testmod(
        optionflags=(doctest.IGNORE_EXCEPTION_DETAIL | doctest.NORMALIZE_WHITESPACE)
    )



from __future__ import annotations

import logging
from typing import Callable, Literal

import numpy as np
import numpy.typing as npt
import scipy.signal

import npc_sessions.utils as utils

logger = logging.getLogger(__name__)

FRAMERATE = 60
"""Visual stim f.p.s - assumed to equal running wheel sampling rate. i.e. one
running wheel sample per camstim vsync"""

RUNNING_SPEED_UNITS: Literal["cm/s", "m/s"] = "m/s"
"""How to report in NWB - NWB expects SI, SDK might have previously reported cm/s"""

RUNNING_LOWPASS_FILTER_HZ = 4
"""Frequency for filtering running speed - filtered data stored in NWB `processing`, unfiltered
in `acquisition`"""


def get_running_speed_from_stim_files(
    *stim_path_or_dataset: utils.StimPathOrDataset,
    sync: utils.SyncPathOrDataset | None = None,
    filt: Callable[[npt.NDArray[np.float64]], npt.NDArray[np.float64]] | None = None,
) -> tuple[npt.NDArray, npt.NDArray]:
    """Pools running speeds across files. Returns arrays of running speed and
    corresponding timestamps."""
    if not sync and len(stim_path_or_dataset) > 1:
        raise ValueError(
            "Must pass sync file to coordinate data from multiple stim files."
        )

    running_speed = np.array([])
    timestamps = np.array([])

    def _append(values, times):
        if len(times) + 1 == len(values):
            values = values[1:]

        if len(times) == len(values):
            times = times[1:]
            values = values[1:]
            times = times + 0.5 * np.median(np.diff(times))
        else:
            raise ValueError(
                f"Length mismatch between running speed ({len(values)}) and timestamps ({len(times)})"
            )
        nonlocal running_speed, timestamps
        # we need to filter before pooling discontiguous blocks of samples
        running_speed = np.concatenate(
            [running_speed, filt(values) if filt else values]
        )
        timestamps = np.concatenate([timestamps, times])

    if sync is None:
        _append(
            get_running_speed_from_hdf5(*stim_path_or_dataset),
            get_frame_times_from_stim_file(*stim_path_or_dataset),
        )
    else:
        # we need timestamps for each frame's nidaq-read time (wheel encoder is read before frame's
        # flip time)
        # there may be multiple h5 files with encoder
        # data per sync file: vsyncs are in blocks with a separating gap
        for hdf5 in stim_path_or_dataset:
            read_times = utils.get_input_data_times(hdf5, sync)
            h5_data = get_running_speed_from_hdf5(hdf5)
            if h5_data is None:
                continue
            _append(h5_data, read_times)

    assert len(running_speed) == len(timestamps)
    return running_speed, timestamps


def get_frame_times_from_stim_file(
    stim_path_or_dataset: utils.StimPathOrDataset,
) -> npt.NDArray:
    return np.concatenate(
        ([0], np.cumsum(utils.get_stim_data(stim_path_or_dataset)["frameIntervals"][:]))
    )


def get_running_speed_from_hdf5(
    stim_path_or_dataset: utils.StimPathOrDataset,
) -> npt.NDArray | None:
    """
    Running speed in m/s or cm/s (see `UNITS`).


    To align with timestamps, remove timestamp[0] and sample[0] and shift
    timestamps by half a frame (speed is estimated from difference between
    timestamps)

    See https://github.com/samgale/DynamicRoutingTask/blob/main/Analysis/DynamicRoutingAnalysisUtils.py
    """
    d = utils.get_stim_data(stim_path_or_dataset)
    if (
        "rotaryEncoder" in d
        and isinstance(d["rotaryEncoder"][()], bytes)
        and d["rotaryEncoder"].asstr()[()] == "digital"
    ):
        if "frameRate" in d:
            assert d["frameRate"][()] == FRAMERATE
        wheel_revolutions = (
            d["rotaryEncoderCount"][:] / d["rotaryEncoderCountsPerRev"][()]
        )
        if not any(wheel_revolutions):
            return None
        wheel_radius_cm = d["wheelRadius"][()]
        if RUNNING_SPEED_UNITS == "m/s":
            running_disk_radius = wheel_radius_cm / 100
        elif RUNNING_SPEED_UNITS == "cm/s":
            running_disk_radius = wheel_radius_cm
        else:
            raise ValueError(
                f"Unexpected units for running speed: {RUNNING_SPEED_UNITS}"
            )
        speed = np.diff(wheel_revolutions * 2 * np.pi * running_disk_radius * FRAMERATE)
        # we lost one sample due to diff: pad with nan to keep same number of samples
        return np.concatenate([[np.nan], speed])
    return None


def lowpass_filter(running_speed: npt.NDArray) -> npt.NDArray:
    """
    Careful not to filter discontiguous blocks of samples.
    See
    https://github.com/AllenInstitute/AllenSDK/blob/36e784d007aed079e3cad2b255ca83cdbbeb1330/allensdk/brain_observatory/behavior/data_objects/running_speed/running_processing.py
    """
    b, a = scipy.signal.butter(
        3, Wn=RUNNING_LOWPASS_FILTER_HZ, fs=FRAMERATE, btype="lowpass"
    )
    return scipy.signal.filtfilt(b, a, np.nan_to_num(running_speed))



from __future__ import annotations

from typing import TYPE_CHECKING

import matplotlib.pyplot as plt

if TYPE_CHECKING:
    import pandas as pd
    import pynwb

    import npc_sessions

import npc_sessions.utils as utils


def plot_unit_quality_metrics_per_probe(session: npc_sessions.DynamicRoutingSession):
    units: pd.DataFrame = session.units[:]

    metrics = [
        "drift_ptp",
        "isi_violations_ratio",
        "amplitude",
        "amplitude_cutoff",
        "presence_ratio",
    ]
    probes = units["device_name"].unique()

    x_labels = {
        "presence_ratio": "fraction of session",
        "isi_violations_ratio": "violation rate",
        "drift_ptp": "microns",
        "amplitude": "uV",
        "amplitude_cutoff": "frequency",
    }

    for metric in metrics:
        fig, ax = plt.subplots(1, len(probes))
        probe_index = 0
        fig.suptitle(f"{metric}")
        for probe in probes:
            units_probe_metric = units[units["device_name"] == probe][metric]
            ax[probe_index].hist(units_probe_metric, bins=20)
            ax[probe_index].set_title(f"{probe}")
            ax[probe_index].set_xlabel(x_labels[metric])
            probe_index += 1

        fig.set_size_inches([10, 6])
    plt.tight_layout()


def plot_all_unit_spike_histograms(session: npc_sessions.DynamicRoutingSession):
    units: pd.DataFrame = session.units[:]

    probes = units["device_name"].unique()

    for probe in probes:
        fig, ax = plt.subplots()
        unit_spike_times = units[units["device_name"] == probe][
            "spike_times"
        ].to_numpy()

        hist, bins = utils.bin_spike_times(unit_spike_times, bin_time=1)
        ax.plot(hist)
        ax.set_title(f"{probe} Spike Histogram")
        ax.set_xlabel("Time (s)")
        ax.set_ylabel("Spike Count per 1 second bin")


def plot_unit_spikes_channels(
    session: npc_sessions.DynamicRoutingSession,
    lower_channel: int = 0,
    upper_channel: int = 384,
):
    units: pd.DataFrame = session.units[:]

    probes = units["device_name"].unique()
    for probe in probes:
        fig, ax = plt.subplots()
        unit_spike_times = units[units["device_name"] == probe]
        unit_spike_times_channel = unit_spike_times[
            (unit_spike_times["peak_channel"] >= lower_channel)
            & (unit_spike_times["peak_channel"] <= upper_channel)
        ]["spike_times"].to_numpy()
        hist, bins = npc_sessions.bin_spike_times(unit_spike_times_channel, bin_time=1)

        ax.plot(hist)
        ax.set_title(
            f"{probe} spike hist for channel range {lower_channel} to {upper_channel}"
        )
        ax.set_xlabel("Time (s)")
        ax.set_ylabel("Spike Count per 1 second bin")


def plot_drift_maps(
    session: npc_sessions.DynamicRoutingSession | pynwb.NWBFile,
) -> tuple[plt.Figure, ...]:
    figs = []
    for _k, v in session.analysis["drift_maps"]:
        fig, ax = plt.subplots()
        ax.imshow(v)
        ax.set_title(f"{session.session_id}")
        ax.axis("off")
        fig.set_size_inches([8, 8])
        fig.tight_layout()
        figs.append(fig)
    return tuple(figs)



"""
Helper functions for getting paths from spikeinterface output (developed for use
with the aind kilosort 2.5 "pipeline" spike-sorting capsule).
"""

from __future__ import annotations

import dataclasses
import functools
import io
import json
from typing import Union

import npc_lims
import npc_session
import numpy as np
import numpy.typing as npt
import pandas as pd
import upath
from typing_extensions import TypeAlias

import npc_sessions.utils as utils

SpikeInterfaceData: TypeAlias = Union[
    str, npc_session.SessionRecord, utils.PathLike, "SpikeInterfaceKS25Data"
]


def get_spikeinterface_data(
    session_or_root_path: SpikeInterfaceData,
) -> SpikeInterfaceKS25Data:
    """Return a SpikeInterfaceKS25Data object for a session.

    >>> paths = get_spikeinterface_data('668759_20230711')
    >>> paths.root == get_spikeinterface_data('s3://codeocean-s3datasetsbucket-1u41qdg42ur9/4797cab2-9ea2-4747-8d15-5ba064837c1c').root
    True
    """
    if isinstance(session_or_root_path, SpikeInterfaceKS25Data):
        return session_or_root_path
    try:
        session = npc_session.SessionRecord(str(session_or_root_path))
        root = None
    except ValueError:
        session = None
        root = utils.from_pathlike(session_or_root_path)
    return SpikeInterfaceKS25Data(session=session, root=root)


@dataclasses.dataclass(unsafe_hash=True, eq=True)
class SpikeInterfaceKS25Data:
    """The root directory of the result data asset produced by the 'pipeline'
    KS2.5 sorting capsule contains `processing.json`, `postprocessed`,
    `spikesorted`, etc. This class just simplifies access to the data in those
    files and dirs.

    Provide a session ID or a root path:
    >>> paths = SpikeInterfaceKS25Data('668759_20230711')
    >>> paths.root
    S3Path('s3://codeocean-s3datasetsbucket-1u41qdg42ur9/4797cab2-9ea2-4747-8d15-5ba064837c1c')

    >>> paths.template_metrics_dict('probeA')
    {'metric_names': ['peak_to_valley', 'peak_trough_ratio', 'half_width', 'repolarization_slope', 'recovery_slope'], 'sparsity': None, 'peak_sign': 'neg', 'upsampling_factor': 10, 'window_slope_ms': 0.7}

    >>> paths.quality_metrics_df('probeA').columns
    Index(['num_spikes', 'firing_rate', 'presence_ratio', 'snr',
           'isi_violations_ratio', 'isi_violations_count', 'rp_contamination',
           'rp_violations', 'sliding_rp_violation', 'amplitude_cutoff',
           'drift_ptp', 'drift_std', 'drift_mad', 'isolation_distance', 'l_ratio',
           'd_prime'],
          dtype='object')
    """

    session: str | npc_session.SessionRecord | None = None
    root: upath.UPath | None = None

    def __post_init__(self) -> None:
        if self.root is None and self.session is None:
            raise ValueError("Must provide either session or root")
        if self.root is None:
            self.root = npc_lims.get_sorted_data_paths_from_s3(self.session)[0].parent

    @staticmethod
    def format_path(*path_components: utils.PathLike) -> upath.UPath:
        """SpikeInterface makes paths with '#' in them, which is not allowed in s3
        paths in general - run paths through this function to fix them."""
        return utils.from_pathlike("/".join(str(path) for path in path_components))

    @staticmethod
    def read_json(path: upath.UPath) -> dict:
        return json.loads(path.read_text())

    @staticmethod
    def read_csv(path: upath.UPath) -> pd.DataFrame:
        return pd.read_csv(path, index_col=0)

    def get_json(self, filename: str) -> dict:
        assert self.root is not None
        return self.read_json(self.format_path(self.root, filename))

    def get_path(self, dirname: str, probe: str | None) -> upath.UPath:
        """Return a path to a single dir or file: either `self.root/dirname` or, if `probe` is specified,
        the probe-specific sub-path within `self.root/dirname`."""
        assert self.root is not None
        if probe is None:
            return self.format_path(self.root, dirname)
        else:
            return next(
                path
                for path in self.format_path(self.root, dirname).iterdir()
                if npc_session.ProbeRecord(probe).lower() in path.as_posix().lower()
            )

    # json data
    processing_json = functools.partialmethod(get_json, "processing.json")
    subject_json = functools.partialmethod(get_json, "subject.json")
    data_description_json = functools.partialmethod(get_json, "data_description.json")
    procedures_json = functools.partialmethod(get_json, "procedures.json")
    visualization_output_json = functools.partialmethod(
        get_json, "visualization_output.json"
    )

    # dirs
    drift_maps = functools.partialmethod(get_path, "drift_maps")
    output = functools.partialmethod(get_path, "output")
    postprocessed = functools.partialmethod(get_path, "postprocessed")
    sorting_precurated = functools.partialmethod(get_path, "sorting_precurated")
    spikesorted = functools.partialmethod(get_path, "spikesorted")

    @functools.cache
    def quality_metrics_dict(self, probe: str) -> dict:
        return self.read_json(
            self.format_path(
                self.postprocessed(probe), "quality_metrics", "params.json"
            )
        )

    @functools.cache
    def quality_metrics_df(self, probe: str) -> pd.DataFrame:
        return self.read_csv(
            self.format_path(
                self.postprocessed(probe), "quality_metrics", "metrics.csv"
            )
        )

    @functools.cache
    def template_metrics_dict(self, probe: str) -> dict:
        return self.read_json(
            self.format_path(
                self.postprocessed(probe), "template_metrics", "params.json"
            )
        )

    @functools.cache
    def template_metrics_df(self, probe: str) -> pd.DataFrame:
        return self.read_csv(
            self.format_path(
                self.postprocessed(probe), "template_metrics", "metrics.csv"
            )
        )

    @functools.cache
    def templates_average(self, probe: str) -> npt.NDArray[np.floating]:
        return np.load(
            io.BytesIO(
                self.format_path(
                    self.postprocessed(probe), "templates_average.npy"
                ).read_bytes()
            )
        )

    @functools.cache
    def templates_std(self, probe: str) -> npt.NDArray[np.floating]:
        return np.load(
            io.BytesIO(
                self.format_path(
                    self.postprocessed(probe), "templates_std.npy"
                ).read_bytes()
            )
        )

    @functools.cache
    def sorting_cached(self, probe: str) -> dict[str, npt.NDArray]:
        return np.load(
            io.BytesIO(
                self.format_path(
                    self.spikesorted(probe), "sorting_cached.npz"
                ).read_bytes()
            ),
            allow_pickle=True,
        )

    @functools.cache
    def default_qc(self, probe: str) -> npt.NDArray[np.floating]:
        return np.load(
            io.BytesIO(
                self.format_path(
                    self.sorting_precurated(probe), "properties", "default_qc.npy"
                ).read_bytes()
            )
        )

    @functools.cache
    def unit_locations(self, probe: str) -> npt.NDArray[np.floating]:
        return np.load(
            io.BytesIO(
                self.format_path(
                    self.postprocessed(probe), "unit_locations", "unit_locations.npy"
                ).read_bytes()
            )
        )

    @functools.cache
    def sorting_json(self, probe: str) -> dict:
        return self.read_json(
            self.format_path(self.postprocessed(probe), "sorting.json")
        )

    @functools.cache
    def electrode_locations_xy(self, probe: str) -> npt.NDArray[np.floating]:
        return np.array(
            self.sorting_json(probe)["annotations"]["__sorting_info__"]["recording"][
                "properties"
            ]["location"]
        )


if __name__ == "__main__":
    import doctest

    import dotenv

    dotenv.load_dotenv(dotenv.find_dotenv(usecwd=True))
    doctest.testmod(
        optionflags=(doctest.IGNORE_EXCEPTION_DETAIL | doctest.NORMALIZE_WHITESPACE)
    )



from __future__ import annotations

from typing import Any, Callable

bad_tag = "[bold red]"
good_tag = "[bold green]"
cautious_tag = "[bold yellow]"


def bad_string(base: str) -> str:
    return bad_tag + base + bad_tag


def good_string(base: str) -> str:
    return good_tag + base + good_tag


def cautious_string(base: str) -> str:
    return cautious_tag + base + cautious_tag


def determine_string_valence(
    value: Any, good_criterion: bool, bad_criterion: bool
) -> Callable[[str], str]:
    if good_criterion:
        return good_string

    elif bad_criterion:
        return bad_string

    else:
        return cautious_string


def add_valence_to_string(
    basestring: str, value: float | int, good_criterion: bool, bad_criterion: bool
) -> str:
    valence_func = determine_string_valence(value, good_criterion, bad_criterion)

    return valence_func(basestring)



"""Tools for working with Open Ephys raw data files."""
from __future__ import annotations

import doctest
import io
import json
import logging
import re
from collections.abc import Generator, Iterable, Sequence
from typing import Any, Literal, NamedTuple

import numpy as np
import numpy.typing as npt
import upath
import zarr

import npc_sessions.utils as utils

logger = logging.getLogger(__name__)

DEFAULT_PROBES = "ABCDEF"


def get_sync_messages_data(
    sync_messages_path: utils.PathLike,
) -> dict[str, dict[Literal["start", "rate"], int]]:
    """
    Start Time for Neuropix-PXI (107) - ProbeA-AP @ 30000 Hz: 210069564
    Start Time for Neuropix-PXI (107) - ProbeA-LFP @ 2500 Hz: 17505797
    Start Time for NI-DAQmx (109) - PXI-6133 @ 30000 Hz: 210265001

    >>> path = 's3://aind-ephys-data/ecephys_670248_2023-08-03_12-04-15/ecephys_clipped/Record Node 102/experiment1/recording1/sync_messages.txt'
    >>> dirname_to_sample = get_sync_messages_data(path)
    >>> dirname_to_sample['NI-DAQmx-105.PXI-6133']
    {'start': 257417001, 'rate': 30000}
    """

    def label(line) -> str:
        return "".join(
            line.split("Start Time for ")[-1]
            .split(" @")[0]
            .replace(") - ", ".")
            .replace(" (", "-")
        )

    def start(line) -> int:
        return int(line.strip(" ").split("Hz:")[-1])

    def rate(line) -> int:
        return int(line.split("@ ")[-1].split(" Hz")[0])

    return {
        label(line): {
            "start": start(line),
            "rate": rate(line),
        }
        for line in utils.from_pathlike(sync_messages_path).read_text().splitlines()[1:]
    }


class EphysTimingInfoOnPXI(NamedTuple):
    name: str
    continuous: upath.UPath
    """Abs path to device's folder within raw data/continuous/"""
    events: upath.UPath
    """Abs path to device's folder within raw data/events/"""
    ttl: upath.UPath
    """Abs path to device's folder within events/"""
    compressed: upath.UPath | None
    """Abs path to device's zarr storage within ecephys_compressed/, or None if not found"""
    start_sample: int
    """Start sample reported in sync_messages.txt"""
    sampling_rate: float
    """Nominal sample rate reported in sync_messages.txt"""
    ttl_sample_numbers: npt.NDArray
    """Sample numbers on open ephys clock, after subtracting first sample reported in
    sync_messages.txt"""
    ttl_states: npt.NDArray
    """Contents of ttl/states.npy"""


class EphysTimingInfoOnSync(NamedTuple):
    name: str
    device: EphysTimingInfoOnPXI
    """Info with paths"""
    sampling_rate: float
    """Sample rate assessed on the sync clock"""
    start_time: float
    """First sample time (sec) relative to the start of the sync clock"""


def read_array_range_from_npy(path: utils.PathLike, _range: int | slice) -> npt.NDArray:
    """Read specific range without downloading entire array. For 1-D array only, currently."""
    if not isinstance(_range, slice):
        _range = slice(_range, _range + 1)
    path = utils.from_pathlike(path)
    ver_major = int.from_bytes(path.fs.read_bytes(path, start=6, end=7), "little")
    header_len_stop = 10 if ver_major == 1 else 12
    header_len = int.from_bytes(
        path.fs.read_bytes(path, start=8, end=header_len_stop), "little"
    )
    array_start = header_len_stop + header_len
    header_bytes = path.fs.read_bytes(path, start=header_len_stop, end=array_start)
    header = eval(header_bytes.decode("utf-8").strip("\n").strip())
    assert len(header["shape"]) == 1, "Currently supporting 1-D array only"
    dtype = header["descr"]
    num_bytes_per_value = np.dtype(dtype).itemsize
    return np.frombuffer(
        path.fs.read_bytes(
            path,
            start=array_start + _range.start * num_bytes_per_value,
            end=array_start + _range.stop * num_bytes_per_value,
        ),
        dtype=dtype,
    )


def get_ephys_timing_on_pxi(
    recording_dirs: Iterable[utils.PathLike],
    only_devices_including: str | None = None,
) -> Generator[EphysTimingInfoOnPXI, None, None]:
    """
    >>> path = upath.UPath('s3://aind-ephys-data/ecephys_670248_2023-08-03_12-04-15/ecephys_clipped/Record Node 102/experiment1/recording1')
    >>> next(get_ephys_timing_on_pxi(path)).sampling_rate
    30000
    """
    if not isinstance(recording_dirs, Iterable):
        recording_dirs = (recording_dirs,)

    for recording_dir in recording_dirs:
        recording_dir = utils.from_pathlike(recording_dir)
        device_to_sync_messages_data = get_sync_messages_data(
            recording_dir / "sync_messages.txt"
        )  # includes name of each input device used (probe, nidaq)
        for device in device_to_sync_messages_data:
            if (
                only_devices_including
                and only_devices_including.lower() not in device.lower()
            ):
                continue
            continuous = recording_dir / "continuous" / device
            if not continuous.exists():
                continue
            events = recording_dir / "events" / device
            ttl = next(events.glob("TTL*"))

            first_sample_from_continuous_sample_numbers = read_array_range_from_npy(
                continuous / "sample_numbers.npy", 0
            ).item()
            first_sample_from_sync_messages = device_to_sync_messages_data[device][
                "start"
            ]
            if (
                first_sample_from_continuous_sample_numbers
                != first_sample_from_sync_messages
            ):
                logger.debug(
                    f"{first_sample_from_sync_messages =} != {first_sample_from_continuous_sample_numbers =}. This may be due to Record Nodes being out-of-sync (green indicator in GUI). Using value from sample_numbers.npy"
                )
            first_sample_on_ephys_clock = first_sample_from_continuous_sample_numbers

            sampling_rate = device_to_sync_messages_data[device]["rate"]
            ttl_sample_numbers = (
                np.load(io.BytesIO((ttl / "sample_numbers.npy").read_bytes()))
                - first_sample_on_ephys_clock
            )
            ttl_states = np.load(io.BytesIO((ttl / "states.npy").read_bytes()))
            try:
                compressed = clipped_path_to_compressed(continuous)
            except ValueError:
                logger.info(f"No compressed data found for {continuous}")
                compressed = None
            yield EphysTimingInfoOnPXI(
                name=device,
                continuous=continuous,
                events=events,
                ttl=ttl,
                compressed=compressed,
                start_sample=first_sample_on_ephys_clock,
                sampling_rate=sampling_rate,
                ttl_sample_numbers=ttl_sample_numbers,
                ttl_states=ttl_states,
            )


def clipped_path_to_compressed(path: utils.PathLike) -> upath.UPath:
    """
    >>> path = upath.UPath('s3://aind-ephys-data/ecephys_670248_2023-08-03_12-04-15/ecephys_clipped/Record Node 102/experiment1/recording1/continuous/Neuropix-PXI-100.ProbeA-AP')
    >>> clipped_path_to_compressed(path)
    S3Path('s3://aind-ephys-data/ecephys_670248_2023-08-03_12-04-15/ecephys_compressed/experiment1_Record Node 102#Neuropix-PXI-100.ProbeA-AP.zarr')
    """
    path = utils.from_pathlike(path)
    if "ecephys_clipped" not in path.as_posix():
        raise ValueError(f'Expected path to contain "ecephys_clipped", got {path}')
    experiment_re = re.search(r".*(experiment\d+)", path.as_posix())
    record_node_re = re.search(r".*(Record Node \d+)", path.as_posix())
    # /recording?/ isn't part of compressed path: assumes 1 recording per folder, or concats multiple recordings
    if not (experiment_re and record_node_re):
        raise ValueError(f"Could not parse experiment and record node from {path}")
    experiment, record_node = experiment_re.groups()[0], record_node_re.groups()[0]
    device_re = re.match(
        rf".*{record_node}/{experiment}/recording\d+/[^/]+/(.*)", path.as_posix()
    )
    if not device_re:
        raise ValueError(f"Could not parse device from {path}")
    compressed_name = f"{experiment}_{record_node}#{device_re.groups()[0]}.zarr"
    root_path = next(p for p in path.parents if p.name == "ecephys_clipped")
    # cannot construct S3Path de novo from a string including `#`, but we can return
    # the actual path that exists
    return next(
        path
        for path in root_path.with_name("ecephys_compressed").iterdir()
        if path.name == compressed_name
    )


def get_pxi_nidaq_data(
    *recording_dirs: utils.PathLike,
    device_name: str | None = None,
) -> npt.NDArray[np.int16]:
    """
    -channel_idx: 0-indexed
    - if device_name not specified, first and only (assumed) NI-DAQ will be used

    ```
    speaker_channel, mic_channel = 1, 3
    ```

    >>> path = upath.UPath('s3://aind-ephys-data/ecephys_670248_2023-08-03_12-04-15/ecephys_clipped/Record Node 102/experiment1/recording1')
    >>> get_pxi_nidaq_data(path).shape
    (142823472, 8)
    """
    if device_name:
        device = next(
            get_ephys_timing_on_pxi(recording_dirs, only_devices_including=device_name)
        )
    else:
        device = get_pxi_nidaq_device(recording_dirs)
    if device.compressed:
        data = zarr.open(device.compressed, mode="r")
        return data["traces_seg0"]
    else:
        device_metadata = next(
            (
                _
                for _ in get_merged_oebin_file(
                    next(utils.from_pathlike(p).glob("*.oebin")) for p in recording_dirs
                )["continuous"]
                if device.name in _["folder_name"]
            ),
            None,
        )
        if device_metadata is None:
            raise ValueError(
                f"Could not find device metadata for {device.name}: looked for `structure.oebin` files in {recording_dirs}"
            )
        num_channels: int = device_metadata["num_channels"]

        if not (protocol := device.continuous.protocol) or protocol == "file":
            # local file we can memory-map
            dat = np.load(device.continuous / "continuous.dat", mmap_mode="r")
        else:
            logger.warning(
                f"Reading entirety of uncompressed OpenEphys NI-DAQ data from {protocol}. If you only need part of this data, consider using `read_array_range_from_npy` with the path instead."
            )
            dat = np.frombuffer(
                (device.continuous / "continuous.dat").read_bytes(), dtype=np.int16
            )
        return np.reshape(dat, (int(dat.size / num_channels), -1))


def get_pxi_nidaq_device(
    recording_dir: Iterable[utils.PathLike],
) -> EphysTimingInfoOnPXI:
    """NI-DAQmx device info

    >>> path = upath.UPath('s3://aind-ephys-data/ecephys_670248_2023-08-03_12-04-15/ecephys_clipped/Record Node 102/experiment1/recording1')
    >>> get_pxi_nidaq_device(path).ttl.parent.name
    'NI-DAQmx-105.PXI-6133'
    """
    device = tuple(
        get_ephys_timing_on_pxi(recording_dir, only_devices_including="NI-DAQmx-")
    )
    if not device:
        raise FileNotFoundError(
            f"No */continuous/NI-DAQmx-*/ dir found in {recording_dir = }"
        )
    if device and len(device) != 1:
        raise FileNotFoundError(
            f"Expected a single NI-DAQmx folder to exist, but found: {[d.continuous for d in device]}"
        )
    return device[0]


def get_ephys_timing_on_sync(
    sync: utils.SyncPathOrDataset,
    recording_dirs: Iterable[utils.PathLike] | None = None,
    devices: Iterable[EphysTimingInfoOnPXI] | None = None,
    only_devices_including: str | None = None,
) -> Generator[EphysTimingInfoOnSync, None, None]:
    """
    One of `recording_dir` or `devices` must be supplied:
        - all devices in `recording_dir` will be returned
        - or just those specified in `devices`
        (use `get_ephys_timing_on_pxi()` to get a filtered iterable of devices
        in a recording_dir)

    >>> path = upath.UPath('s3://aind-ephys-data/ecephys_670248_2023-08-03_12-04-15/ecephys_clipped/Record Node 102/experiment1/recording1')
    >>> sync = upath.UPath('s3://aind-ephys-data/ecephys_670248_2023-08-03_12-04-15/behavior/20230803T120415.h5')
    >>> device = next(get_ephys_timing_on_sync(sync, path))
    >>> device.sampling_rate, device.start_time
    (30000.070518634246, 20.080209634424037)
    """
    if not (recording_dirs or devices):
        raise ValueError("Must specify recording_dir or devices")

    sync = utils.get_sync_data(sync)

    sync_barcode_times, sync_barcode_ids = utils.extract_barcodes_from_times(
        on_times=sync.get_rising_edges("barcode_ephys", units="seconds"),
        off_times=sync.get_falling_edges("barcode_ephys", units="seconds"),
    )
    if devices and not isinstance(devices, Iterable):
        devices = (devices,)

    if recording_dirs and not isinstance(recording_dirs, Iterable):
        recording_dirs = (recording_dirs,)

    if recording_dirs and not devices:
        devices = get_ephys_timing_on_pxi(recording_dirs, only_devices_including)

    assert devices is not None
    for device in devices:
        (
            ephys_barcode_times,
            ephys_barcode_ids,
        ) = utils.extract_barcodes_from_times(
            on_times=device.ttl_sample_numbers[device.ttl_states > 0]
            / device.sampling_rate,
            off_times=device.ttl_sample_numbers[device.ttl_states < 0]
            / device.sampling_rate,
        )

        timeshift, sampling_rate, _ = utils.get_probe_time_offset(
            master_times=sync_barcode_times,
            master_barcodes=sync_barcode_ids,
            probe_times=ephys_barcode_times,
            probe_barcodes=ephys_barcode_ids,
            acq_start_index=0,
            local_probe_rate=device.sampling_rate,
        )
        start_time = -timeshift
        if (np.isnan(sampling_rate)) | (~np.isfinite(sampling_rate)):
            sampling_rate = device.sampling_rate

        yield EphysTimingInfoOnSync(
            name=device.name,
            device=device,
            sampling_rate=sampling_rate,
            start_time=start_time,
        )


def is_new_ephys_folder(path: utils.PathLike) -> bool:
    """Look for any hallmarks of a v0.6.x Open Ephys recording in path or subfolders."""
    path = utils.from_pathlike(path)
    globs = (
        "Record Node*",
        "structure*.oebin",
    )
    components = tuple(_.replace("*", "") for _ in globs)

    if any(_.lower() in path.as_posix().lower() for _ in components):
        return True

    for glob in globs:
        if next(path.rglob(glob), None):
            return True
    return False


def is_complete_ephys_folder(path: utils.PathLike) -> bool:
    """Look for all hallmarks of a complete v0.6.x Open Ephys recording."""
    # TODO use structure.oebin to check for completeness
    path = utils.from_pathlike(path)
    if not is_new_ephys_folder(path):
        return False
    for glob in ("continuous.dat", "spike_times.npy", "spike_clusters.npy"):
        if not next(path.rglob(glob), None):
            logger.debug(f"Could not find {glob} in {path}")
            return False
    return True


def is_valid_ephys_folder(
    path: utils.PathLike,
    min_size_gb: int | float | None = None,
) -> bool:
    """Check a single dir of raw data for size and v0.6.x+ Open Ephys."""
    path = utils.from_pathlike(path)
    if not path.is_dir():
        return False
    if not is_new_ephys_folder(path):
        return False
    if min_size_gb is not None and utils.dir_size(path) < min_size_gb * 1024**3:  # GB
        return False
    return True


def get_ephys_root(path: utils.PathLike) -> upath.UPath:
    """Returns the parent of the first `Record Node *` found in the supplied
    path.

    >>> get_ephys_root(upath.UPath('A:/test/Record Node 0/Record Node test')).as_posix()
    'A:/test'
    """
    path = utils.from_pathlike(path)
    if "Record Node" not in path.as_posix():
        raise ValueError(
            f"Could not find 'Record Node' in {path} - is this a valid raw ephys path?"
        )
    return next(
        p.parent for p in path.parents if "Record Node".lower() in p.name.lower()
    )


def get_filtered_ephys_paths_relative_to_record_node_parents(
    toplevel_ephys_path: utils.PathLike,
) -> Generator[tuple[upath.UPath, upath.UPath], None, None]:
    """For restructuring the raw ephys data in a session folder, we want to
    discard superfluous recording folders and only keep the "good" data, but
    with the overall directory structure relative to `Record Node*` folders intact.

    Supply a top-level path that contains `Record Node *`
    subfolders somewhere in its tree.

    Returns a generator akin to `path.rglob('Record Node*')` except:
    - only paths associated with the "good" ephys data are returned (with some
    assumptions made about the ephys session)
    - a tuple of two paths is supplied:
        - `(abs_path, abs_path.relative_to(record_node.parent))`
        ie. path[1] should always start with `Record Node *`

    Expectation is:
    - `/npexp_path/ephys_*/Record Node */ recording1 / ... / continuous.dat`

    ie.
    - one recording per `Record Node *` folder
    - one subfolder in `npexp_path/` per `Record Node *` folder (following the
    pipeline `*_probeABC` / `*_probeDEF` convention for recordings split across two
    drives)


    Many folders have:
    - `/npexp_path/Record Node */ recording*/ ...`

    ie.
    - multiple recording folders per `Record Node *` folder
    - typically there's one proper `recording` folder: the rest are short,
    aborted recordings during setup

    We filter out the superfluous small recording folders here.

    Some folders (Templeton) have:
    - `/npexp_path/Record Node */ ...`

    ie.
    - contents of expected ephys subfolders directly deposited in npexp_path

    """
    toplevel_ephys_path = utils.from_pathlike(toplevel_ephys_path)
    record_nodes = toplevel_ephys_path.rglob("Record Node*")

    for record_node in record_nodes:
        superfluous_recording_dirs = tuple(
            _.parent for _ in get_superfluous_oebin_paths(record_node)
        )
        logger.debug(
            f"Found {len(superfluous_recording_dirs)} superfluous recording dirs to exclude: {superfluous_recording_dirs}"
        )

        for abs_path in record_node.rglob("*"):
            is_superfluous_path = any(
                _ in abs_path.parents for _ in superfluous_recording_dirs
            )

            if is_superfluous_path:
                continue

            yield abs_path, abs_path.relative_to(record_node.parent)


def get_raw_ephys_subfolders(
    path: utils.PathLike, min_size_gb: int | float | None = None
) -> tuple[upath.UPath, ...]:
    """
    Return raw ephys recording folders, defined as the root that Open Ephys
    records to, e.g. `A:/1233245678_366122_20220618_probeABC`.
    """
    path = utils.from_pathlike(path)

    subfolders = set()

    for f in upath.UPath(path).rglob("continuous.dat"):
        if any(
            k in f.as_posix().lower()
            for k in [
                "sorted",
                "extracted",
                "curated",
            ]
        ):
            # skip sorted/extracted folders
            continue

        subfolders.add(get_ephys_root(f))

    if min_size_gb is not None:
        subfolders = {
            _ for _ in subfolders if utils.dir_size(_) < min_size_gb * 1024**3
        }

    return tuple(sorted(subfolders, key=lambda s: str(s)))


# - If we have probeABC and probeDEF raw data folders, each one has an oebin file:
#     we'll need to merge the oebin files and the data folders to create a single session
#     that can be processed in parallel
def get_single_oebin_path(path: utils.PathLike) -> upath.UPath:
    """Get the path to a single structure.oebin file in a folder of raw ephys data.

    - There's one structure.oebin per `recording*` folder
    - Raw data folders may contain multiple `recording*` folders
    - Datajoint expects only one structure.oebin file per Session for sorting
    - If we have multiple `recording*` folders, we assume that there's one
        good folder - the largest - plus some small dummy / accidental recordings
    """
    path = utils.from_pathlike(path)
    if not path.is_dir():
        raise ValueError(f"{path} is not a directory")

    oebin_paths = tuple(path.rglob("structure*.oebin"))

    if not oebin_paths:
        raise FileNotFoundError(f"No structure.oebin file found in {path}")

    if len(oebin_paths) == 1:
        return oebin_paths[0]

    oebin_parents = (_.parent for _ in oebin_paths)
    dir_sizes = tuple(utils.dir_size(_) for _ in oebin_parents)
    return oebin_paths[dir_sizes.index(max(dir_sizes))]


def get_superfluous_oebin_paths(path: utils.PathLike) -> tuple[upath.UPath, ...]:
    """Get the paths to any oebin files in `recording*` folders that are not
    the largest in a folder of raw ephys data.

    Companion to `get_single_oebin_path`.
    """
    path = utils.from_pathlike(path)
    all_oebin_paths = tuple(path.rglob("structure*.oebin"))

    if len(all_oebin_paths) == 1:
        return ()

    return tuple(set(all_oebin_paths) - {(get_single_oebin_path(path))})


def assert_xml_files_match(*path: utils.PathLike) -> None:
    """Check that all xml files are identical, as they should be for
    recordings split across multiple locations e.g. A:/*_probeABC, B:/*_probeDEF
    or raise an error.

    Update: xml files on two nodes can be created at slightly different times, so their `date`
    fields may differ. Everything else should be identical.
    """
    paths = tuple(utils.from_pathlike(p) for p in path)
    if not all(s == ".xml" for s in [p.suffix for p in paths]):
        raise ValueError(f"Not all paths are XML files: {paths}")
    if not all(p.is_file() for p in paths):
        raise FileNotFoundError(
            f"Not all paths are files, or they do not exist: {paths}"
        )
    if not utils.checksums_match(*paths):
        data = [utils.get_settings_xml_data(p) for p in paths]
        if not all(d == data[0] for d in data):
            raise AssertionError(f"XML files do not match: {paths}")


def get_merged_oebin_file(
    paths: Iterable[utils.PathLike], exclude_probe_names: Sequence[str] | None = None
) -> dict[Literal["continuous", "events", "spikes"], list[dict[str, Any]]]:
    """Merge two or more structure.oebin files into one.

    For recordings split across multiple locations e.g. A:/*_probeABC,
    B:/*_probeDEF
    - if items in the oebin files have 'folder_name' values that match any
    entries in `exclude_probe_names`, they will be removed from the merged oebin
    """
    if not isinstance(paths, Iterable):
        paths = tuple(paths)
    oebin_paths = tuple(utils.from_pathlike(p) for p in paths)
    if len(oebin_paths) == 1:
        return read_oebin(oebin_paths[0])

    # ensure oebin files can be merged - if from the same exp they will have the same settings.xml file
    if any(p.suffix != ".oebin" for p in oebin_paths):
        raise ValueError(f"Not all paths are .oebin files: {oebin_paths}")
    assert_xml_files_match(
        *[p / "settings.xml" for p in [o.parent.parent.parent for o in oebin_paths]]
    )

    logger.debug(f"Creating merged oebin file from {oebin_paths}")
    merged_oebin: dict = {}
    for oebin_path in sorted(oebin_paths):
        oebin_data = read_oebin(oebin_path)

        for key in oebin_data:
            # skip if already in merged oebin
            if merged_oebin.get(key) == oebin_data[key]:
                continue

            # 'continuous', 'events', 'spikes' are lists, which we want to concatenate across files
            if isinstance(oebin_data[key], list):
                for item in oebin_data[key]:
                    # skip if already in merged oebin
                    if item in merged_oebin.get(key, []):
                        continue

                    # skip probes in excl list (ie. not inserted)
                    if exclude_probe_names and any(
                        p.lower() in item.get("folder_name", "").lower()
                        for p in exclude_probe_names
                    ):
                        continue

                    # insert in merged oebin
                    merged_oebin.setdefault(key, []).append(item)

    if not merged_oebin:
        raise ValueError("No data found in structure.oebin files")
    return merged_oebin


def read_oebin(
    path: utils.PathLike,
) -> dict[Literal["continuous", "events", "spikes"], list[dict[str, Any]]]:
    return json.loads(utils.from_pathlike(path).read_text())


if __name__ == "__main__":
    import doctest

    import dotenv

    dotenv.load_dotenv(dotenv.find_dotenv(usecwd=True))
    doctest.testmod(
        optionflags=(doctest.IGNORE_EXCEPTION_DETAIL | doctest.NORMALIZE_WHITESPACE)
    )



"""
>>> path = upath.UPath('s3://aind-ephys-data/ecephys_668759_2023-07-11_13-07-32/ecephys_clipped/Record Node 102/settings.xml')
>>> et = ET.parse(io.BytesIO(path.read_bytes()))
>>> _hostname(et)
'W10DT05516'
>>> _date_time(et)
(datetime.date(2023, 7, 11), datetime.time(13, 7, 53))
>>> _open_ephys_version(et)
'0.6.4'
>>> _settings_xml_md5(path)
'5c1b33293cb7c5f72df56fbdd4b72fc3'
>>> _probe_serial_numbers(et)
(18194810652, 18005123131, 18005102491, 19192719021, 18005118602, 19192719061)
>>> _probe_letters(et)
('A', 'B', 'C', 'D', 'E', 'F')
>>> _probe_types(et)
('Neuropixels 1.0', 'Neuropixels 1.0', 'Neuropixels 1.0', 'Neuropixels 1.0', 'Neuropixels 1.0', 'Neuropixels 1.0')
>>> isinstance(get_settings_xml_data(path), SettingsXmlInfo)
True
>>> _probe_serial_number_to_channel_pos_xy(et)[18194810652]['CH0']
(27, 0)
"""

from __future__ import annotations

import dataclasses
import datetime
import doctest
import functools
import hashlib
import io
import xml.etree.ElementTree as ET
from typing import Literal

import upath

import npc_sessions.utils as utils


@dataclasses.dataclass
class SettingsXmlInfo:
    """Info from a settings.xml file from an Open Ephys recording."""

    path: upath.UPath
    probe_serial_numbers: tuple[int, ...]
    probe_types: tuple[str, ...]
    probe_letters: tuple[Literal["A", "B", "C", "D", "E", "F"], ...]
    hostname: str
    date: datetime.date
    start_time: datetime.time
    open_ephys_version: str
    channel_pos_xy: tuple[dict[str, tuple[int, int]], ...]

    def __eq__(self, other) -> bool:
        if not isinstance(other, SettingsXmlInfo):
            return NotImplemented
        # files from multiple nodes can be created at slightly different times, so their `date`
        # fields may differ. Everything else should be identical.
        return all(
            getattr(self, field.name) == getattr(other, field.name)
            for field in dataclasses.fields(self)
            if field.name not in ("path", "start_time")
        )


def get_settings_xml_etree(path: utils.PathLike) -> ET.ElementTree:
    """Info from a settings.xml file from an Open Ephys recording."""
    return ET.parse(io.BytesIO(utils.from_pathlike(path).read_bytes()))


def get_settings_xml_data(path: utils.PathLike | SettingsXmlInfo) -> SettingsXmlInfo:
    if isinstance(path, SettingsXmlInfo):
        return path
    et = get_settings_xml_etree(path)
    return SettingsXmlInfo(
        path=utils.from_pathlike(path),
        probe_serial_numbers=_probe_serial_numbers(et),
        probe_types=_probe_types(et),
        probe_letters=_probe_letters(et),
        hostname=_hostname(et),
        date=_date_time(et)[0],
        start_time=_date_time(et)[1],
        open_ephys_version=_open_ephys_version(et),
        channel_pos_xy=tuple(_probe_serial_number_to_channel_pos_xy(et).values()),
    )


def _get_tag_text(et: ET.ElementTree, tag: str) -> str | None:
    result = [
        element.text for element in et.getroot().iter() if element.tag == tag.upper()
    ]
    if not (result and any(result)):
        result = [element.attrib.get(tag.lower()) for element in et.getroot().iter()]
    return str(result[0]) if (result and any(result)) else None


def _get_tag_attrib(et: ET.ElementTree, tag: str, attrib: str) -> str | None:
    result = [
        element.attrib.get(attrib)
        for element in et.getroot().iter()
        if element.tag == tag.upper()
    ]
    return str(result[0]) if (result and any(result)) else None


def _hostname(et: ET.ElementTree) -> str:
    result = (
        # older, pre-0.6.x:
        _get_tag_text(et, "machine")
        # newer, 0.6.x:
        or _get_tag_attrib(et, "MACHINE", "name")
    )
    if not result:
        raise LookupError(f"No hostname: {result!r}")
    return result


@functools.cache
def _date_time(et: ET.ElementTree) -> tuple[datetime.date, datetime.time]:
    """Date and recording start time."""
    result = _get_tag_text(et, "date")
    if not result:
        raise LookupError(f"No datetime found: {result!r}")
    dt = datetime.datetime.strptime(result, "%d %b %Y %H:%M:%S")
    return dt.date(), dt.time()


@functools.cache
def _probe_attrib_dicts(et: ET.ElementTree) -> tuple[dict[str, str], ...]:
    return tuple(
        probe_dict.attrib
        for probe_dict in et.getroot().iter()
        if "probe_serial_number" in probe_dict.attrib
    )


def _probe_attrib(et: ET.ElementTree, attrib: str) -> tuple[str, ...]:
    return tuple(probe[attrib] for probe in _probe_attrib_dicts(et))


def _probe_serial_numbers(et: ET.ElementTree) -> tuple[int, ...]:
    return tuple(int(_) for _ in _probe_attrib(et, "probe_serial_number"))


def _probe_types(et: ET.ElementTree) -> tuple[str, ...]:
    try:
        return _probe_attrib(et, "probe_name")
    except KeyError:
        return tuple("unknown" for _ in _probe_attrib_dicts(et))


def _probe_serial_number_to_channel_pos_xy(
    et: ET.ElementTree,
) -> dict[int, dict[str, tuple[int, int]]]:
    x_pos_iter = tuple(a for a in et.getroot().iter() if a.tag == "ELECTRODE_XPOS")
    y_pos_iter = tuple(a for a in et.getroot().iter() if a.tag == "ELECTRODE_YPOS")

    return dict(
        zip(
            _probe_serial_numbers(et),
            (
                {k: (int(x.attrib[k]), int(y.attrib[k])) for k in x.attrib.keys()}
                for x, y in zip(x_pos_iter, y_pos_iter)
            ),
        )
    )


def _probe_idx(et: ET.ElementTree) -> tuple[int, ...]:
    """Try to reconstruct index from probe slot and port.

    assumptions:
        - 2 slots
        - 3 ports in use per slot
        - probes ABCDEF connected in sequence
    """
    slots, ports = _probe_attrib(et, "slot"), _probe_attrib(et, "port")
    result = tuple(
        (int(s) - int(min(slots))) * len(set(ports)) + int(p) - 1
        for s, p in zip(slots, ports)
    )
    if not all(idx in range(6) for idx in result):
        raise ValueError(f"probe_idx: {result!r}, slots: {slots}, ports: {ports}")
    return result


def _probe_letters(
    et: ET.ElementTree,
) -> tuple[Literal["A", "B", "C", "D", "E", "F"], ...]:
    return tuple("ABCDEF"[idx] for idx in _probe_idx(et))  # type: ignore [misc]


def _open_ephys_version(et: ET.ElementTree) -> str:
    result = _get_tag_text(et, "version")
    if not result:
        raise LookupError(f"No version found: {result!r}")
    return result


def _settings_xml_md5(path: str | upath.UPath) -> str:
    return hashlib.md5(upath.UPath(path).read_bytes()).hexdigest()


if __name__ == "__main__":
    import doctest

    import dotenv

    dotenv.load_dotenv(dotenv.find_dotenv(usecwd=True))
    doctest.testmod(
        optionflags=(doctest.IGNORE_EXCEPTION_DETAIL | doctest.NORMALIZE_WHITESPACE)
    )



from npc_sessions.plots.audio import *
from npc_sessions.plots.behavior import *
from npc_sessions.plots.spikes import *
from npc_sessions.plots.sync import *
from npc_sessions.plots.timing import *
from npc_sessions.plots.video import *
